# flake8: noqa: E501
from datetime import date, datetime

from sqlalchemy.schema import Column
from sqlalchemy.types import Date, DateTime, Integer, String

from sqltask.classes import dq
from sqltask.classes.exceptions import TooFewRowsException
from sqltask.classes.sql import LookupSource, SqlDataSource
from sqltask.classes.table import DqTableContext

from .base_task import BaseExampleTask


class FactCustomerTask(BaseExampleTask):
    def __init__(self, report_date: date):
        super().__init__(report_date=report_date)

        # Define the metadata for the main fact table
        self.add_table(DqTableContext(
            name="fact_customer",
            engine_context=self.ENGINE_TARGET,
            columns=[
                Column("report_date", Date, comment="Date of snapshot", primary_key=True),
                Column("etl_timestamp", DateTime, comment="Timestamp when row was created", nullable=False),
                Column("customer_name", String(10), comment="Unique customer identifier (name)", primary_key=True),
                Column("birthdate", Date, comment="Birthdate of customer if defined and in the past", nullable=True),
                Column("age", Integer, comment="Age of customer in years if birthdate defined", nullable=True),
                Column("sector_code", String(10), comment="Sector code of customer", nullable=True),
            ],
            comment="The customer table",
            timestamp_column_name="etl_timestamp",
            batch_params={"report_date": report_date},
            dq_info_column_names=["etl_timestamp"],
        ))

        # Define the main query used to populate the target table
        self.add_data_source(SqlDataSource(
            name="main",
            sql="""
SELECT name,
       birthday
FROM customers
WHERE report_date = :report_date
            """,
            params={"report_date": report_date},
            engine_context=self.ENGINE_SOURCE,
        ))

        # Define a lookup source used for enriching the main source query
        self.add_lookup_source(LookupSource(
            name="sector_code",
            sql="""
SELECT name,
       sector_code
FROM sector_codes
WHERE start_date <= :report_date
  AND end_date > :report_date
            """,
            params={"report_date": report_date},
            engine_context=self.ENGINE_SOURCE,
        ))

    def transform(self) -> None:
        report_date = self.batch_params["report_date"]
        sector_code_lookup = self.get_lookup("sector_code")
        for in_row in self.get_data_source("main"):
            row = self.get_new_row("fact_customer")

            # customer_name
            customer_name = in_row["name"]
            row["customer_name"] = customer_name

            # birthdate
            birthday = in_row["birthday"]
            age = None
            try:
                birthdate = datetime.strptime(birthday, "%Y-%m-%d").date() if birthday else None
                age = None
                if birthdate is None:
                    row.log_dq(
                        column_name="birthdate",
                        source=dq.Source.SOURCE,
                        priority=dq.Priority.MEDIUM,
                        category=dq.Category.MISSING,
                        message="Missing birthdate",
                    )
                elif birthdate > report_date:
                    row.log_dq(
                        column_name="birthdate",
                        source=dq.Source.SOURCE,
                        priority=dq.Priority.HIGH,
                        category=dq.Category.INCORRECT,
                        message=f"Birthdate in future: {birthday}",
                    )
                    birthdate = None
                else:
                    age = int((report_date - birthdate).days / 365.25)
            except ValueError:
                # parse error
                row.log_dq(
                    column_name="birthdate",
                    source=dq.Source.SOURCE,
                    priority=dq.Priority.HIGH,
                    category=dq.Category.INCORRECT,
                    message=f"Cannot parse birthdate: {birthday}"
                )
                birthdate = None
            row["birthdate"] = birthdate

            # age
            if age is None:
                row.log_dq(
                    column_name="age",
                    source=dq.Source.TRANSFORM,
                    priority=dq.Priority.MEDIUM,
                    category=dq.Category.MISSING,
                    message="Age is undefined due to undefined birthdate",
                )
            row["age"] = age

            # sector_code
            sector_code = sector_code_lookup.get(customer_name)
            if sector_code is None:
                row.log_dq(
                    column_name="sector_code",
                    source=dq.Source.SOURCE,
                    priority=dq.Priority.LOW,
                    category=dq.Category.MISSING,
                    message="Sector code undefined in lookup table"
                )
            row["sector_code"] = sector_code

            # Finally add row to table output
            row.append()

    def validate(self):
        if len(self.get_table_context("fact_customer").output_rows) < 2:
            raise TooFewRowsException("There should never be less than 2 rows")



import logging
import os
from typing import Optional

from sqltask.classes.table import TableContext
from sqltask.engine_specs.base import BaseEngineSpec, UploadType
from sqltask.utils.engine_specs import create_tmp_csv


class BigQueryEngineSpec(BaseEngineSpec):
    engine = 'bigquery'
    supports_column_comments = True
    supports_table_comments = True
    supports_schemas = False

    @classmethod
    def _insert_rows_csv(cls, table_context: "TableContext") -> None:
        cls._insert_rows(table_context, UploadType.CSV)

    @classmethod
    def _insert_rows(cls, table_context: TableContext,
                     upload_type: Optional[UploadType] = None) -> None:
        """
        BigQuery bulk loading is done by exporting the data to CSV and using the
        `google-cloud-bigquery` library to upload data.

        :param table_context: the target table to upload into
        :param upload_type: If undefined, defaults to whichever ´UploadType` is defined
        in `default_upload_type`.
        """
        from google.cloud import bigquery

        file_path = create_tmp_csv(table_context)
        client = bigquery.Client()
        database = table_context.engine_context.engine.url.database
        schema = table_context.schema
        table_id = table_context.table.name

        # dataset_ref = client.dataset(file_path)
        # table_ref = dataset_ref.table(table_id)

        job_config = bigquery.LoadJobConfig()
        job_config.source_format = bigquery.SourceFormat.CSV
        job_config.skip_leading_rows = 0
        job_config.autodetect = True

        table_id = f"{database}.{schema}.{table_id}"

        try:
            with open(file_path, 'r', encoding="utf-8", newline='') as csv_file:
                job = client.load_table_from_file(
                    csv_file, table_id, job_config=job_config)
        finally:
            job.result()
            logging.info(f"Loaded {job.output_rows} rows into {table_id}")
            os.remove(f"{file_path}")



from enum import Enum


class Priority(Enum):
    MANDATORY = "mandatory"
    HIGH = "high"
    MEDIUM = "medium"
    LOW = "low"


class Source(Enum):
    SOURCE = "source"
    TRANSFORM = "transform"
    LOOKUP = "lookup"


class Category(Enum):
    MISSING = "missing"
    INCORRECT = "incorrect"
    DUPLICATE = "duplicate"



class SqlTaskException(Exception):
    pass


class MandatoryValueMissingException(SqlTaskException):
    pass


class TooFewRowsException(SqlTaskException):
    pass


class ExecutionArgumentException(SqlTaskException):
    pass



from collections import UserDict
from typing import Any, Dict, Mapping, NamedTuple, Optional

from sqltask.classes.engine import EngineContext
from sqltask.classes.table import TableContext


class QueryContext(NamedTuple):
    sql: str
    params: Dict[str, Any]
    table_context: Optional[TableContext]
    engine_context: EngineContext


class BaseDataSource:
    def __init__(self, name: str):
        self.name = name

    def __iter__(self):
        raise NotImplementedError("`__iter__` not implemented")


class Lookup(UserDict):
    def __init__(self, data_source: BaseDataSource, kv: Mapping[Any, Any]):
        super().__init__(kv)
        self.data_source = data_source



import inspect
import pkgutil
from importlib import import_module
from pathlib import Path
from typing import Dict, Type

from sqltask.engine_specs.base import BaseEngineSpec

_engines: Dict[str, Type[BaseEngineSpec]] = {}

for (_, name, _) in pkgutil.iter_modules([Path(__file__).parent]):  # type: ignore
    imported_module = import_module("." + name, package=__name__)

    for i in dir(imported_module):
        attribute = getattr(imported_module, i)

        if (
            inspect.isclass(attribute)
            and issubclass(attribute, BaseEngineSpec)
            and attribute.engine != ""
        ):
            _engines[attribute.engine] = attribute


def get_engine_spec(engine_name: str) -> Type[BaseEngineSpec]:
    """
    Get an engine spec based on an engine name, e.g. snowflake.

    :param engine_name: Name of engine, i.e. engine.name
    :return: Engine spec for a given engine name. Returns `BaseEngineSpec`
    if engine does not have a dedicated spec available.
    """
    return _engines.get(engine_name, BaseEngineSpec)



# flake8: noqa: E501
import os
from datetime import datetime

from sqlalchemy.schema import Column
from sqlalchemy.types import Date, String

from sqltask.classes.file import CsvDataSource
from sqltask.classes.table import TableContext

from .base_task import BaseExampleTask


class InitSourceTask(BaseExampleTask):
    def __init__(self):
        super().__init__()
        current_dir = os.path.dirname(__file__)

        # main customer table
        self.add_table(TableContext(
            name="customers",
            engine_context=self.ENGINE_SOURCE,
            columns=[
                Column("report_date", Date, comment="Monthly snapshot date", primary_key=True),  # noqa
                Column("name", String(128), comment="Customer name", primary_key=True),
                Column("birthday", String(10), comment="Birthdate of customer in unreliable yyyy-mm-dd string format", nullable=True),  # noqa
            ],
            comment="The original customer table",
        ))
        # csv file containing data
        self.add_data_source(CsvDataSource(
            name="customers.csv",
            file_path=os.path.join(current_dir, "..", "static_files", "customers.csv"),
            delimiter=",",
        ))

        # additional table with sector codes per customer
        self.add_table(TableContext(
            name="sector_codes",
            engine_context=self.ENGINE_SOURCE,
            columns=[
                Column("start_date", Date, comment="date when row becomes active (inclusive)", nullable=False),
                Column("end_date", Date, comment="date when row becomes inactive (non-inclusive)", nullable=False),
                Column("name", String(128), comment="Customer name (non-unique)", nullable=False),
                Column("sector_code", String(10), comment="Sector code of cutomer", nullable=True),
            ],
            comment="Sector codes for customers",
        ))
        # csv file containing data
        self.add_data_source(CsvDataSource(
            name="sector_codes.csv",
            file_path=os.path.join(current_dir, "..", "static_files", "sector_codes.csv"),
            delimiter=",",
        ))

    def transform(self) -> None:
        # populate customers table
        for in_row in self.get_data_source("customers.csv"):
            row = self.get_new_row("customers")
            row["report_date"] = datetime.strptime(in_row["report_date"], "%Y-%m-%d").date()
            row["name"] = in_row["name"]
            row["birthday"] = in_row["birthday"]
            row.append()

        # populate sector_codes table
        for in_row in self.get_data_source("sector_codes.csv"):
            row = self.get_new_row("sector_codes")
            row["start_date"] = datetime.strptime(in_row["start_date"], "%Y-%m-%d").date()
            row["end_date"] = datetime.strptime(in_row["end_date"], "%Y-%m-%d").date()
            row["name"] = in_row["name"]
            row["sector_code"] = in_row["sector_code"]
            row.append()



import os

from sqltask.classes.table import TableContext
from sqltask.engine_specs.base import BaseEngineSpec, UploadType
from sqltask.utils.engine_specs import create_tmp_csv


class PostgresEngineSpec(BaseEngineSpec):
    engine = 'postgresql'
    supported_uploads = (UploadType.SQL_INSERT,
                         UploadType.SQL_INSERT_MULTIROW,
                         UploadType.CSV,
                         )
    default_upload_type = UploadType.CSV
    supports_column_comments = True
    supports_table_comments = True
    supports_schemas = True

    @classmethod
    def _insert_rows_csv(cls, table_context: TableContext) -> None:
        """
        Postgres bulk loading is done by exporting the data to CSV and using the
        cursor `copy_from` method.

        :param table_context: the target table to upload into
        """
        file_path = create_tmp_csv(table_context)
        engine = table_context.engine_context.engine
        columns = [column.name for column in table_context.table.columns]
        engine.raw_connection()
        try:
            with engine.connect() as conn:
                cursor = conn.connection.cursor()
                with open(file_path, 'r', encoding="utf-8", newline='') as csv_file:
                    cursor.copy_from(file=csv_file,
                                     table=table_context.table.name,
                                     columns=columns,
                                     null="")
                    csv_file.close()
                    conn.connection.commit()
        finally:
            os.remove(f"{file_path}")



import logging
from collections import UserDict
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import sqlalchemy as sa
from sqlalchemy.schema import Column, Table
from sqlalchemy.types import String

from sqltask.classes import dq

if TYPE_CHECKING:
    from sqltask.classes.engine import EngineContext


class TableContext:
    def __init__(
            self,
            name: str,
            engine_context: "EngineContext",
            columns: List[Column],
            comment: Optional[str] = None,
            schema: Optional[str] = None,
            batch_params: Optional[Dict[str, Any]] = None,
            timestamp_column_name: Optional[str] = None,
            table_params: Dict[str, Any] = None,
    ):
        """
        Create a new table context.

        :param name: Name of target table in database.
        :param engine_context: engine to bind table to.
        :param columns: All columns in table.
        :param comment: Table comment.
        :param schema: Schema to use. If left unspecified, falls back to the schema
        provided by the engine context
        :param batch_params: Mapping between column names and values that are used to
        delete old rows in the table.
        :param timestamp_column_name: Name of column used for populating etl timestamp.
        """
        # comment is apparently not Optional, so needs to be passed via kwargs
        table_params = table_params or {}
        if comment:
            table_params["comment"] = comment
        table = Table(name,
                      engine_context.metadata,
                      *columns,
                      **table_params)

        # Finalize main table context after dq table context is created
        self.name = name
        self.table = table
        self.engine_context = engine_context
        self.batch_params = batch_params or {}
        self.timestamp_column_name = timestamp_column_name
        self.schema = schema or engine_context.schema
        self.output_rows: List[Dict[str, Any]] = []

    def get_new_row(self) -> "OutputRow":
        """
        Get a new row intended to be added to the table.
        """
        output_row = OutputRow(self)
        if self.timestamp_column_name:
            output_row[self.timestamp_column_name] = datetime.utcnow()
        return output_row

    def delete_rows(self) -> None:
        """
        Delete old rows from target table that match batch parameters.
        """
        self.engine_context.engine_spec.truncate_rows(self)

    def insert_rows(self) -> None:
        """
        Insert rows into target tables.
        """
        self.engine_context.engine_spec.insert_rows(self)

    def migrate_schema(self) -> None:
        """
        Migrate table schema to correspond to table definition.
        """
        table = self.table
        engine = self.engine_context.engine
        metadata = self.engine_context.metadata
        if engine.has_table(table.name, schema=self.schema):
            # table exists, add column
            inspector = sa.inspect(engine)
            cols_existing = [col['name'] for col in inspector.get_columns(table.name)]
            for column in table.columns:
                if column.name not in cols_existing:
                    self.engine_context.engine_spec.add_column(self, column)

            # remove redundant columns
            cols_new = {col.name: col for col in table.columns}
            for column_name in cols_existing:
                if column_name not in cols_new:
                    self.engine_context.engine_spec.drop_column(self, column_name)
        else:
            # table doesn't exist, create new
            logging.debug(f"Create new table `{table.name}`")
            metadata.create_all(tables=[table])


class DqTableContext(TableContext):
    """
    A table context with the ability to log data quality issues
    """
    def __init__(
            self,
            name: str,
            engine_context: "EngineContext",
            columns: List[Column],
            comment: Optional[str] = None,
            schema: Optional[str] = None,
            batch_params: Optional[Dict[str, Any]] = None,
            timestamp_column_name: Optional[str] = None,
            table_params: Dict[str, Any] = None,
            dq_table_name: Optional[str] = None,
            dq_engine_context: Optional["EngineContext"] = None,
            dq_schema: Optional[str] = None,
            dq_info_column_names: Optional[List[str]] = None,
            dq_table_params: Dict[str, Any] = None,
    ):
        """
        :param name: Name of target table in database.
        :param engine_context: engine to bind table to.
        :param columns: All columns in table.
        :param comment: Table comment.
        :param schema: Schema to use. If left unspecified, falls back to the schema
        provided by the engine context
        :param batch_params: Mapping between column names and values that are used to
        delete old rows in the table.
        :param timestamp_column_name: Name of column used for populating etl timestamp.
        :param table_params: Additional parameters to be passed as kwargs to the
        Table constructor.
        :param dq_table_name: Name of data quality table. Defaults to original table
        name + `_dq`
        :param dq_engine_context: Engine context used for creating data quality table.
        Defaults to `engine_context` if omitted
        :param dq_schema: Schema used for creating data quality table. Defaults to
        `schema` if left blank
        :param dq_info_column_names: Name of columns to be appended to the data
        quality table for informational purposes that aren't primary keys, e.g.
        customer name.
        :param table_params: Additional parameters to be passed as kwargs to the
        data quality Table constructor.
        """
        super().__init__(
            name=name,
            engine_context=engine_context,
            columns=columns,
            comment=comment,
            schema=schema,
            batch_params=batch_params,
            timestamp_column_name=timestamp_column_name,
            table_params=table_params,
        )

        dq_table_name = dq_table_name or self.name + "_dq"
        dq_engine_context = dq_engine_context or self.engine_context
        dq_timestamp_column_name = self.timestamp_column_name or "etl_timestamp"
        dq_schema = dq_schema or self.schema
        dq_info_column_names = dq_info_column_names or []
        dq_table_params = dq_table_params or {}

        # primary key columns
        batch_columns = []
        primary_key_columns = []
        info_columns = []

        for column in self.table.columns:
            # make copy of column and remove primary keys as one row might
            # have several data quality issues. Also make nullable, as data quality
            # issues don't necessarily relate to a specific column in the target table
            column_copy = column.copy()
            column_copy.primary_key = False
            column_copy.nullable = True
            if column.name in self.batch_params:
                batch_columns.append(column_copy)
            elif column.primary_key:
                primary_key_columns.append(column_copy)
            elif column.name in dq_info_column_names:
                info_columns.append(column_copy)

        default_dq_columns = [
            Column("source", String, comment="Source of issue"),
            Column("priority", String, comment="Priority of issue"),
            Column("category", String, comment="Category of issue"),
            Column("column_name", String, comment="Affected column in target table"),
            Column("message", String, comment="Verbose description of the issue"),
        ]

        dq_columns = batch_columns + primary_key_columns + info_columns + \
            default_dq_columns

        dq_schema = dq_schema or self.schema
        self.dq_table_context = TableContext(
            name=dq_table_name,
            engine_context=dq_engine_context,
            columns=dq_columns,
            comment=comment,
            schema=dq_schema,
            batch_params=batch_params,
            timestamp_column_name=dq_timestamp_column_name,
            table_params=dq_table_params,
        )

    def delete_rows(self) -> None:
        """
        Delete old rows from target table that match batch parameters.
        """
        super().delete_rows()
        self.dq_table_context.delete_rows()

    def insert_rows(self) -> None:
        """
        Insert rows into target tables.
        """
        super().insert_rows()
        self.dq_table_context.insert_rows()

    def migrate_schema(self) -> None:
        """
        Migrate table schema to correspond to table definition.
        """
        super().migrate_schema()
        self.dq_table_context.migrate_schema()


class OutputRow(UserDict):
    def __init__(self, table_context: TableContext):
        super().__init__(table_context.batch_params)
        self.table_context = table_context
        if table_context.timestamp_column_name:
            self[table_context.timestamp_column_name] = datetime.utcnow()

    def log_dq(self, column_name: Optional[str], category: dq.Category,
               priority: dq.Priority, source: dq.Source,
               message: Optional[str] = None):
        """
        Log data quality issue to be recorded in data quality table.

        :param column_name: Name of affected column in target table.
        :param category: The type of data quality issue.
        :param source: To what phase the data quality issue relates.
        :param priority: What the priority of the data quality issue is.
        Should be None for aggregate data quality issues.
        :param message: Verbose description of observed issue.
        """
        if column_name not in self.table_context.table.columns:
            raise Exception(f"Column `{column_name}` not in table "
                            f"`{self.table_context.table.name}`")

        if not isinstance(self.table_context, DqTableContext):
            raise Exception("table is not instance of DqTableContext")

        dq_table_context = self.table_context.dq_table_context
        if dq_table_context is None:
            raise Exception(f"No dq table context defined for {self.table_context.name}")
        dq_output_row = dq_table_context.get_new_row()

        dq_output_row.update({
            "source": str(source.value),
            "priority": str(priority.value),
            "category": str(category.value),
            "column_name": column_name,
            "message": message
        })

        # make sure all keys from dq table are included in final row object
        dq_row = {}
        for column in dq_table_context.table.columns:
            # Add additional info columns from main row
            if column.name not in dq_output_row and column.name in self.keys():
                dq_output_row[column.name] = self[column.name]
            elif column.name not in dq_output_row:
                raise Exception(f"No column `{column.name}` in output row for table "
                                f"`{dq_output_row.table_context.table.name}`")
            dq_row[column.name] = dq_output_row[column.name]
        dq_table_context.output_rows.append(dq_row)

    def append(self) -> None:
        """
        Append the row to the table.
        """

        output_row = {}
        for column in self.table_context.table.columns:
            if column.name not in self:
                raise Exception(f"No column `{column.name}` in output row for table "
                                f"`{self.table_context.name}`")
            output_row[column.name] = self[column.name]
        self.table_context.output_rows.append(output_row)



import logging
from collections import namedtuple
from typing import TYPE_CHECKING, Any, Dict, Optional

from sqlalchemy.sql import text

from sqltask.classes.common import BaseDataSource, Lookup

if TYPE_CHECKING:
    from sqltask.classes.engine import EngineContext


class SqlDataSource(BaseDataSource):
    def __init__(
            self,
            name: str,
            sql: str,
            params: Dict[str, Any],
            engine_context: "EngineContext",
            schema: str = None,
    ):
        """
        :param name: name of data source
        :param sql: sql query with parameter values prefixed with a colon, e.g.
        `WHERE dt <= :batch_date`
        :param params: mapping between parameter keys and values, e.g.
        `{"batch_date": date(2010, 1, 1)}`
        :param schema: schema to use when executing query. Uses schema defined
        in sql_engine if left undefined.
        :param engine_context: engine used to execute the query.
        """
        params = params or {}
        schema = schema or engine_context.schema
        super().__init__(name)
        self.sql = sql
        self.params = params or {}
        self.engine_context = engine_context
        self.schema = schema or engine_context.schema

    def __iter__(self):
        rows = self.engine_context.engine.execute(text(self.sql), self.params)
        for row in rows:
            yield Lookup(self, dict(zip(row.keys(), row.values())))


class LookupSource(BaseDataSource):
    def __init__(self,
                 name: str,
                 sql: str,
                 params: Dict[str, Any],
                 engine_context: "EngineContext",
                 schema: Optional[str] = None,
                 keys: int = 1,
                 ):
        """
        :param name: name of data source
        :param sql: sql query with parameter values prefixed with a colon, e.g.
        `WHERE dt <= :batch_date`
        :param params: mapping between parameter keys and values, e.g.
        `{"batch_date": date(2010, 1, 1)}`
        :param schema: schema to use when executing query. Uses schema defined
        in sql_engine if left undefined.
        :param keys: number of keys in dict key.
        :param engine_context: engine used to execute the query.
        """
        super().__init__(name)
        self.sql = sql
        self.params = params or {}
        self.engine_context = engine_context
        self.schema = schema or engine_context.schema
        self.keys = keys

    def get_lookup(self) -> Lookup:
        rows = self.engine_context.engine.execute(text(self.sql), self.params)
        cursor = rows.cursor
        self.KeyTuple = namedtuple(self.name, rows.keys())
        if self.keys < 1:
            raise Exception(f"A minimum of 1 key is needed for a lookup")
        elif len(rows.keys()) <= self.keys:
            raise Exception(
                f"Too few columns in lookup `name`: {len(cursor.description)} "
                f"found, expected more than {self.keys}")
        else:
            lookup: Dict[str, Any] = {}

        row_count, duplicate_count = 0, 0
        for row in rows:
            row_count += 1
            # regular key-value lookup
            if self.keys == 1:
                key = row[0]
            else:
                key = self.KeyTuple([row.values()])

            if len(rows.keys()) == self.keys + 1:
                value = row[self.keys]
            else:
                value = {rows.keys()[i]: row[i]
                         for i in range(self.keys, len(rows.keys()))}
                value = Lookup(self, value)
            if key in lookup:
                duplicate_count += 1
            else:
                lookup[key] = value

        if duplicate_count > 0:
            logging.warning(
                f"Query result for lookup `{self.name}` has {duplicate_count} "
                f"duplicate keys, ignoring duplicate rows")
        logging.info(f"Finished populating lookup `{self.name}` with {len(lookup)} rows")

        return Lookup(self, lookup)



import logging
from typing import Optional

from sqlalchemy.engine import create_engine
from sqlalchemy.schema import MetaData

from sqltask.engine_specs import get_engine_spec


class EngineContext:
    def __init__(self,
                 name: str,
                 url: str,
                 schema: Optional[str] = None,
                 **kwargs):
        self.name = name
        self.engine = create_engine(url)
        self.engine_spec = get_engine_spec(self.engine.name)
        self.schema = schema or self.engine_spec.get_schema_name(self.engine.url)
        self.metadata = MetaData(bind=self.engine, schema=self.schema, **kwargs)
        logging.info(f"Created engine `{name}` using "
                     f"`{self.engine_spec.__name__}` on schema `{schema}`")



import os

from sqltask.classes.table import TableContext
from sqltask.engine_specs.base import BaseEngineSpec, UploadType
from sqltask.utils.engine_specs import create_tmp_csv


class MssqlEngineSpec(BaseEngineSpec):
    engine = 'mssql'
    supported_uploads = (UploadType.SQL_INSERT,
                         UploadType.SQL_INSERT_MULTIROW,
                         )
    default_upload_type = UploadType.SQL_INSERT_MULTIROW
    supports_column_comments = True
    supports_table_comments = True
    supports_schemas = True

    @classmethod
    def _insert_rows_csv(cls, table_context: TableContext) -> None:
        """
        MSSQL bulk loading is done by exporting the data to CSV and using the
        cursor `copy_from` method.

        :param table_context: the target table to upload into
        """
        file_path = create_tmp_csv(table_context)
#        engine = table_context.engine_context.engine
#        columns = [column.name for column in table_context.table.columns]

        with table_context.engine_context.engine.begin() as conn:
            stmt = f"BULK INSERT {table_context.table.name} FROM '{file_path}' " \
                   f"WITH (FIRSTROW = 2, FIELDTERMINATOR = '\t', ROWTERMINATOR = '\n')"
            conn.execute(stmt)
        os.remove(f"{file_path}")



import os

from sqltask.classes.table import TableContext
from sqltask.engine_specs.base import BaseEngineSpec, UploadType
from sqltask.utils.engine_specs import create_tmp_csv


class SnowflakeEngineSpec(BaseEngineSpec):
    engine = "snowflake"
    supported_uploads = (UploadType.SQL_INSERT,
                         UploadType.SQL_INSERT_MULTIROW,
                         UploadType.CSV,
                         )
    default_upload_type = UploadType.CSV

    @classmethod
    def _insert_rows_csv(cls, table_context: TableContext) -> None:
        """
        Snowflake bulk loading is done by exporting the data to CSV and using the
        PUT + COPY statement to upload the data.

        :param table_context: the target table to upload into
        """
        file_path = create_tmp_csv(table_context)
        table = table_context.table
        engine = table_context.engine_context.engine

        try:
            with engine.connect() as conn:
                conn.execute(f"CREATE OR REPLACE TEMPORARY STAGE {table.name}")
                conn.execute(f"PUT file://{file_path} @{table.name}")
                conn.execute(f"COPY INTO {table.name} FROM @{table.name} "
                             f"FILE_FORMAT = (TYPE = 'CSV' FIELD_DELIMITER = '\t' "
                             f"SKIP_HEADER = 0 EMPTY_FIELD_AS_NULL = TRUE "
                             f"COMPRESSION = GZIP) FORCE = TRUE")
        finally:
            os.remove(f"{file_path}")



from sqlalchemy.schema import Column

from sqltask.classes.table import TableContext
from sqltask.engine_specs.base import BaseEngineSpec, UploadType


class SqliteEngineSpec(BaseEngineSpec):
    engine = 'sqlite'
    default_upload_type = UploadType.SQL_INSERT
    supports_column_comments = False
    supports_table_comments = False
    supports_schemas = False

    @classmethod
    def drop_column(cls,
                    table_context: TableContext,
                    column_name: Column,
                    ) -> None:
        # TODO: sqlite doesn't support dropping columns; implement the workaround from
        # https: // www.techonthenet.com / sqlite / tables / alter_table.php
        pass



import io

from setuptools import find_packages, setup

from sqltask import __version__

with io.open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='sqltask',
    version=__version__,
    description="ETL tool based on SqlAlchemy for building robust ETL pipelies with "
                "high emphasis on high data quality",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/villebro/sqltask',
    author='Ville Brofeldt',
    author_email='villebro@apache.org',
    packages=find_packages(),
    install_requires=[
        'sqlalchemy',
    ],
    license='MIT',
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
    ],
    extras_require={
        "bigquery": ["pybigquery>=0.4.11"],
        "postgres": ["psycopg2>=2.8.3"],
        "mssql": ["pymssql>=2.1.4"],
        "snowflake": ["snowflake-sqlalchemy>=1.1.14"],
    },
)



import logging
from enum import Enum
from typing import Optional, Sequence

from sqlalchemy.engine.url import URL
from sqlalchemy.schema import Column
from sqlalchemy.sql import text

from sqltask.classes.table import TableContext

log = logging


class UploadType(Enum):
    SQL_INSERT = 1
    SQL_INSERT_MULTIROW = 2
    CSV = 3


class BaseEngineSpec:
    """
    Generic spec defining default behaviour for SqlAlchemy engines.
    """
    engine: Optional[str] = None
    default_upload_type = UploadType.SQL_INSERT
    supported_uploads: Sequence[UploadType] = (
        UploadType.SQL_INSERT,
        UploadType.SQL_INSERT_MULTIROW,
    )
    supports_column_comments = True
    supports_table_comments = True
    supports_schemas = True

    @classmethod
    def insert_rows(cls,
                    table_context: "TableContext",
                    upload_type: Optional[UploadType] = None) -> None:
        """
        Default method for inserting data into database. This

        :param output_rows: Rows to upload.
        :param table_context: Table context on which the upload should be based.
        :param upload_type: If undefined, defaults to whichever ´UploadType` is defined
        in `default_upload_type`.
        """
        upload_type = upload_type or cls.default_upload_type
        if upload_type == UploadType.SQL_INSERT:
            cls._insert_rows_sql_insert(table_context)
        elif upload_type == UploadType.SQL_INSERT_MULTIROW:
            cls._insert_rows_sql_insert_multirow(table_context)
        elif upload_type == UploadType.CSV:
            cls._insert_rows_csv(table_context)
        else:
            raise NotImplementedError(f"Unsupported upload type: {upload_type}")

    @classmethod
    def _insert_rows_sql_insert(cls,
                                table_context: "TableContext"
                                ) -> None:
        """
        Insert rows using standard insert statements. Not very performant, but mostly
        universally supported.
        """
        if UploadType.SQL_INSERT not in cls.supported_uploads:
            raise Exception(f"SQL INSERT not supported by `{cls.__name__}`")
        with table_context.engine_context.engine.begin() as conn:
            conn.execute(table_context.table.insert(), *table_context.output_rows)

    @classmethod
    def _insert_rows_sql_insert_multirow(cls,
                                         table_context: "TableContext",
                                         chunksize: int = 5000
                                         ) -> None:
        """
        Insert rows using standard insert statements. Not very performant, but mostly
        universally supported.
        """
        if UploadType.SQL_INSERT not in cls.supported_uploads:
            raise Exception(f"SQL INSERT not supported by `{cls.__name__}`")
        with table_context.engine_context.engine.begin() as conn:
            conn.execute(table_context.table.insert().values(table_context.output_rows))

    @classmethod
    def _insert_rows_csv(cls, table_context: "TableContext") -> None:
        raise NotImplementedError(f"`{cls.__name__}` does not support CSV upload")

    @classmethod
    def truncate_rows(cls, table_context: "TableContext") -> None:
        """
        Delete old rows from target table that match the execution parameters.

        :param table_context: Output table
        """
        table = table_context.table
        engine = table_context.engine_context.engine
        batch_params = table_context.batch_params
        if batch_params:
            where_clause = " WHERE " + " AND ".join(
                [f"{col} = :{col}" for col in batch_params.keys()])
        else:
            where_clause = ""
        stmt = f"DELETE FROM {table.name}{where_clause}"
        engine.execute(text(stmt), batch_params)

    @classmethod
    def get_schema_name(cls, url: URL) -> Optional[str]:
        """
        Extract schema name from URL instance. Assumes that the schema name is what
        cmes after a slash in the database name, e.g. `database/schema`.

        :param url: SqlAlchemy URL instance
        :return: schema name
        """
        schema = None
        database = url.database
        if cls.supports_schemas and database is not None and "/" in database:
            schema = database.split("/")[1]
        return schema

    @classmethod
    def add_column(cls,
                   table_context: TableContext,
                   column: Column,
                   ) -> None:
        """
        Add a column to target table

        :param table_context: table which to alter
        :param column: column to be added
        :return:
        """
        table_name = table_context.table.name
        logging.debug(f"Add column `{column.name}` to table `{table_name}`")
        stmt = f'ALTER TABLE {table_name} ADD COLUMN ' \
               f'{column.name} {str(column.type)}'
        table_context.engine_context.engine.execute(stmt)

    @classmethod
    def drop_column(cls,
                    table_context: TableContext,
                    column_name: Column,
                    ) -> None:
        """
        Add a column to target table

        :param table_context: table which to alter
        :param column_name: column to drop
        :return:
        """
        table_name = table_context.table.name
        logging.debug(f"Drop column `{column_name}` from table `{table_name}`")
        stmt = f'ALTER TABLE {table_name} DROP COLUMN {column_name}'
        table_context.engine_context.engine.execute(stmt)



import logging
import os
from typing import Any, Dict

from sqltask.classes.common import BaseDataSource, Lookup
from sqltask.classes.sql import LookupSource
from sqltask.classes.table import OutputRow, TableContext

__version__ = '0.3.1'

# initialize logging
log = logging.getLogger('sqltask')
log_level = os.getenv("SQLTASK_LOG_LEVEL")
if log_level:
    log.setLevel(log_level)


class SqlTask:
    def __init__(self, **batch_params):
        """
        Main class in library.

        :param batch_params: Mapping from batch column name to value
        """
        self._tables: Dict[str, TableContext] = {}
        self._data_sources: Dict[str, BaseDataSource] = {}
        self._lookup_sources: Dict[str, LookupSource] = {}
        self.batch_params: Dict[str, Any] = batch_params or {}

    def add_table(self, table_context: TableContext) -> None:
        """
        Add a table schema.

        :param table_context: a table context to be added to the task
        """
        self._tables[table_context.name] = table_context

    def get_table_context(self, name: str) -> TableContext:
        """
        Retrieve table context

        :param name: name of table context
        :return: predefined table context
        """
        table_context = self._tables.get(name)
        if table_context is None:
            raise Exception(f"Undefined table context: {name}")
        return table_context

    def transform(self) -> None:
        """
        Main transformation method where target rows should be generated.
        """
        raise NotImplementedError("`transform` method must be implemented")

    def validate(self) -> None:
        """
        Abstract validation method that is executed after transformation is completed.
        Should be implemented to validate aggregate measures that can't be validated
        during transformation.
        """
        pass

    def get_new_row(self, name: str) -> OutputRow:
        """
        Returns an empty row based on the schema of the table.

        :param name: Name of output table.
        :return: An output row prepopulated with batch and etl columns.
        """
        table_context = self._tables.get(name)
        if table_context is None:
            raise Exception(f"Undefined table context: `{name}`")
        return OutputRow(table_context)

    def add_data_source(self, data_source: BaseDataSource) -> None:
        """
        Add a data source that can be iterated over.

        :param data_source: an instance of base class BaseDataSource
        """
        self._data_sources[data_source.name] = data_source

    def add_lookup_source(self, lookup_source: LookupSource) -> None:
        """
        Add a data source that can be iterated over.

        :param lookup_source: an instance of base class LookupSource
        """
        self._lookup_sources[lookup_source.name] = lookup_source

    def get_data_source(self, name: str) -> BaseDataSource:
        """
        Get results for a predefined query.

        :param name: name of query that has been added with the `self.add_source_query`
        method.

        :return: The DataSource instance that can be iterated over
        """
        log.debug(f"Retrieving source query `{name}`")
        data_source = self._data_sources.get(name)
        if data_source is None:
            raise Exception(f"Data source `{data_source}` not found")
        return data_source

    def get_lookup(self, name: str) -> Lookup:
        """
        Get results for a predefined lookup query. The results for are cached when the
        method is called for the first time.

        :param name: name of query that has been added with the `self.add_lookup_query`
        method.

        :return: A lookup, which can be a single or
        """
        lookup = self._lookup_sources.get(name)
        if lookup is None:
            raise Exception(f"Lookup `{name}` not found")
        return lookup.get_lookup()

    def insert_rows(self) -> None:
        """
        Insert rows in target tables.
        """
        for table_context in self._tables.values():
            table_context.insert_rows()

    def delete_rows(self) -> None:
        """
        Delete rows in target tables.
        """
        for table_context in self._tables.values():
            table_context.delete_rows()

    def migrate_schemas(self) -> None:
        """
        Migrate all table schemas to target engines. Create new tables if missing,
        add missing columns if table exists but not all columns present.
        """
        for table_context in self._tables.values():
            table_context.migrate_schema()

    def execute_migration(self):
        log.debug("Start schema migrate")
        self.migrate_schemas()

    def execute_etl(self):
        log.debug(f"Start transform")
        self.transform()
        log.debug(f"Start validate")
        self.validate()
        log.debug(f"Start delete")
        self.delete_rows()
        log.debug(f"Start insert")
        self.insert_rows()
        log.debug(f"Finish etl")

    def execute(self):
        self.execute_migration()
        self.execute_etl()



import csv
from typing import List

from sqltask.classes.common import BaseDataSource, Lookup


class CsvDataSource(BaseDataSource):
    """
    Data source that reads from a CSV file.

    """

    def __init__(self, name: str, file_path: str, delimiter: str):
        """
        :param name: name of data source.
        :param file_path: path to the csv file.
        :param delimiter: csv file delimiter.
        """

        super().__init__(name)
        self.file_path = file_path
        self.delimiter = delimiter

        # populate column names
        self.columns: List[str] = []
        with open(self.file_path, newline="") as csvfile:
            csvreader = csv.reader(csvfile, delimiter=self.delimiter)
            row = next(csvreader)
            for column in row:
                self.columns.append(column)

    def __iter__(self):
        with open(self.file_path, newline="") as csvfile:
            csvreader = csv.reader(csvfile, delimiter=self.delimiter)
            # skip header row
            next(csvreader)
            for in_row in csvreader:
                row_dict = {self.columns[i]: col for i, col in enumerate(in_row)}
                yield Lookup(self, row_dict)



import csv
import logging
import tempfile
from datetime import datetime

from sqltask.classes.table import TableContext


def create_tmp_csv(table_context: TableContext, delimiter: str = "\t") -> str:
    """
    Creates a temporary csv file based on `output_rows`.

    :param table_context: Table context based on which the csv columns will be based.
    :param delimiter: Delmiter to use when exporting csv.
    :return: the path of the created temporary csv file.
    """
    csv_rows = []
    for row in table_context.output_rows:
        csv_row = []
        for column in table_context.table.columns:
            csv_row.append(row[column.name])
        csv_rows.append(csv_row)

    table = table_context.table
    epoch = str(datetime.utcnow().timestamp())
    file_path = f"{tempfile.gettempdir()}/{table.name}_{epoch}.csv"
    logging.info(f"Creating temporary file `{file_path}`")

    with open(file_path, 'w', encoding="utf-8", newline='') as csv_file:
        writer = csv.writer(csv_file, delimiter=delimiter)
        writer.writerows(csv_rows)
    csv_file.close()
    return file_path



import os

from sqltask import SqlTask
from sqltask.classes.engine import EngineContext


class BaseExampleTask(SqlTask):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        source_url = os.getenv("SQLTASK_SOURCE", "sqlite:///source.db")
        target_url = os.getenv("SQLTASK_TARGET", "sqlite:///target.db")
        self.ENGINE_SOURCE = EngineContext("source", source_url)
        self.ENGINE_TARGET = EngineContext("target", target_url)



from datetime import date

from tasks.fact_customer_task import FactCustomerTask
from tasks.init_source_task import InitSourceTask

if __name__ == "__main__":
    # create initial data used by main task
    task = InitSourceTask()
    task.execute()

    # execute main task
    task = FactCustomerTask(report_date=date(2019, 6, 30))
    task.execute()
